// VinciaMatching.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VINCIA's Matching class.
  
#ifndef VINCIAMATCHING_H
#define VINCIAMATCHING_H

#include "Basics.h"                   // Vec4
#include "VinciaAntenna.h"            // Antenna
#include "VinciaEvolutionVariables.h" // EvolutionVariable
#include "VinciaTools.h"              // VinClu
#include "VinciaMGInterface.h"        // MGInterface

using namespace Pythia8;

namespace Vincia {

  class Matching {
    
  public:
    
    // Constructor.
    Matching() {isInit=false; isInitPtr=false;};

    // Destructor.
    virtual ~Matching() {};

    // Initialize Pointers (must be done before init)
    void initPtr(EvolutionVariable* RPtrIn, MGInterface* mgInterfacePtrIn, 
		 Settings* settingsPtrIn, VinClu* vinCluPtrIn) {
      RPtr           = RPtrIn;
      mgInterfacePtr = mgInterfacePtrIn; 
      settingsPtr    = settingsPtrIn; 
      vinCluPtr      = vinCluPtrIn; 
      isInitPtr=true;
    }

    // Initialize Pointers to Antenna Sets
    void initAntPtr(AntennaSet* antV0, AntennaSet* antV1=0, 
		    AntennaSet* antV2=0, AntennaSet* antV3=0) {
      antennaSets.resize(0);
      antennaSets.push_back(antV0);
      if (antV1 != 0) antennaSets.push_back(antV1);
      if (antV2 != 0) antennaSets.push_back(antV2);
      if (antV3 != 0) antennaSets.push_back(antV3);
    }

    // Initialize
    bool init();

    // Main functions to return whether and how far a specific decaying particle
    // is matched
    //   isMatched : return true if any matching exists at this order
    bool isMatched(int idRes, int nEmit, int nSplit, bool tryMG=false);
    //   antTree : return |M_n|^2/|M_Born|^2 for these daughters
    // double antTree(int idRes, vector<Particle> daughters);

    // *********************************************************************
    // Return a matching coefficient
    // *********************************************************************

    bool getPME(int nIn, vector<int> id, vector<Vec4> p, vector<double>& PME);

    // *********************************************************************
    // Return a 2->4 matching coefficient
    // *********************************************************************

    bool getP2to4(vector<int> id, vector<Vec4> p, vector<double>& PME);

    // *********************************************************************
    // Hardcoded Matrix Elements and Antenna Functions
    // *********************************************************************

    // GGG functions 
    
    // 3-parton tree-level GGG antennae
    double A30(double sar,double srb,double sab,double s);
    double sA30(double yar,double yrb);
    double A30(vector<Vec4>&,int,int,int);
    double D30(double sar,double srb,double sab,double s);
    double E30(double sar,double srb,double sab,double s);
    double F30(double sar,double srb, double sab, double s);
    double G30(double sar,double srb,double sab,double s);
    double Q30(double sar,double srb,double s);
    // 3-parton tree-level GGG antennae, singular terms only
    double A30sng(double sar,double srb,double sab,double s);
    double D30sng(double sar,double srb,double sab,double s);
    double E30sng(double sar,double srb,double sab,double s);
    double F30sng(double sar,double srb, double sab, double s);
    double G30sng(double sar,double srb,double sab,double s);
    // 3-parton tree-level Skands antennae
    double E30pzs(double sar,double srb,double sab,double s);
    // 3-parton tree-level default ("Med") VINCIA antennae
    double E30med(double sar,double srb, double sab, double s);
    // 4-parton tree-level GGG antennae
    double A40(vector<Vec4>& pn, int a, int b, int c, int d);
    double A40aux(vector<Vec4>& pn, int a,int b,int c,int d);
    double A40NLC(vector<Vec4>& pn, int a, int b, int c, int d);
    double A40auxNLC(vector<Vec4>& pn, int a,int b,int c,int d);
    double B40(vector<Vec4>& pn, int a, int b, int c, int d);
    double B40aux(vector<Vec4>& pn, int a, int b, int c, int d);
    double C40(vector<Vec4>& pn, int a, int b, int c, int d);
    double C40aux(vector<Vec4>& pn, int a, int b, int c, int d);
    double D40(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40a(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40b(vector<Vec4>& pn, int a, int b, int c, int d);
    double F40(vector<Vec4>& pn, int a, int b, int c, int d);
    double G40(vector<Vec4>& pn, int a, int b, int c, int d);
    double G40NLC(vector<Vec4>& pn, int a, int b, int c, int d);
    double H40(vector<Vec4>& pn, int a, int b, int c, int d);
    // 3-parton 1-loop GGG antennae (PS-subtracted)
    double A31fin(double s12,double s23,double s13,double s123);
    double A31finNLC(double s12,double s23,double s13,double s123);
    double A31auxR(double y, double z);
    
    // *********************************************************************
    // Ariadne functions
    double A30ar(double sar, double srb, double s);
    double D30ar(double sar, double srb, double s);
    double E30ar(double sar, double srb, double s);
    double F30ar(double sar, double srb, double s);
    
    // *********************************************************************
    // Functions imported from the GGG form code, for counter-checks
    // A40LC checked, rel diffs < 1e-15.
    // A40NLC sum (= complete ~A40) checked, rel diffs < 1e-15.
    // B40 checked against form code and against Ellis,Ross,Terrano, d < 1e-13. 
    // C40 checked against form code
    // D40 (above) imported directly from form code
    // E40 checked against form code, rel diffs < 1e-15.
    // H40 checked against form code, rel diffs < 1e-14.
    
    double A40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    double A40NLCfrm(vector<Vec4>& pn, int a, int b, int c, int d);
    double B40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    double C40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40NLCfrm(vector<Vec4>& pn, int a, int b, int c, int d);
    double H40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    
    // *********************************************************************
    // Functions copied from Ellis, Ross, and Terrano
    
    double B40ert(vector<Vec4>& pn, int a, int b, int c, int d);
    
    // *********************************************************************
    // Kosower's sector functions
    
    double Asec30gks(double sar,double srb,double s);
    double Dsec30gks(double sar,double srb,double s);
    double Esec30gks(double sar,double srb,double s);
    double Fsec30gks(double sar,double srb,double s);
    double Gsec30gks(double sar,double srb,double s);
    
    // *********************************************************************
    // Skands' experimental sector functions
    
    double Asec30ska(double sar,double srb,double s);
    double Dsec30ska(double sar,double srb,double s);
    double Esec30ska(double sar,double srb,double s);
    double Fsec30ska(double sar,double srb,double s);
    double Gsec30ska(double sar,double srb,double s);
   
    // Set verbose level
    void setVerbose(int verboseIn) {verbose = verboseIn;}

  private:

    // Is initialized?    
    bool isInit, isInitPtr;

    // Color Depth (0:Leading, 1:Full, 2:Full allow negative)
    int colorDepth;

    // Is MG Matching switched on?
    bool isMGOn;

    // Uncertainty Bands
    bool uncertaintyBands;

    // Pointers to VINCIA objects
    EvolutionVariable* RPtr;
    MGInterface*       mgInterfacePtr;
    Settings*          settingsPtr;
    VinClu*            vinCluPtr;

    // Vector of pointers to antenna sets (0 is physical, >= 1 are variations)
    vector<AntennaSet*> antennaSets;    

    // Ordering Type
    int orderingType;

    // Sector ordering
    bool sectorOrdering;

    // Verbose level
    int verbose;

  };

}

#endif
